from abc import (
    ABCMeta,
)
from typing import (
    Type,
)

from edu_rdm_integration.adapters.helpers import (
    WebEduRunnerHelper,
)
from edu_rdm_integration.adapters.results import (
    WebEduRunnerResult,
)
from edu_rdm_integration.adapters.validators import (
    WebEduRunnerValidator,
)
from function_tools.runners import (
    BaseRunner,
    GlobalHelperRunner,
)


class WebEduRunner(BaseRunner, metaclass=ABCMeta):
    """
    Базовый класс ранеров функций продуктов Образования.
    """

    def _prepare_helper_class(self) -> Type[WebEduRunnerHelper]:
        """
        Возвращает класс помощника ранера функции.
        """
        return WebEduRunnerHelper

    def _prepare_validator_class(self) -> Type[WebEduRunnerValidator]:
        """
        Возвращает класс валидатора ранера функции.
        """
        return WebEduRunnerValidator

    def _prepare_result_class(self) -> Type[WebEduRunnerResult]:
        """
        Возвращает класс результата ранера функции.
        """
        return WebEduRunnerResult


class WebEduGlobalHelperRunner(GlobalHelperRunner, metaclass=ABCMeta):
    """
    Базовый класс для создания ранеров выполнения запускаемых объектов с
    глобальным помощником продуктов Образования.
    """

    def _prepare_helper_class(self) -> Type[WebEduRunnerHelper]:
        """
        Возвращает класс помощника ранера функции.
        """
        return WebEduRunnerHelper

    def _prepare_validator_class(self) -> Type[WebEduRunnerValidator]:
        """
        Возвращает класс валидатора ранера функции.
        """
        return WebEduRunnerValidator

    def _prepare_result_class(self) -> Type[WebEduRunnerResult]:
        """
        Возвращает класс результата ранера функции.
        """
        return WebEduRunnerResult



from edu_rdm_integration.management.general import (
    BaseExportEntityDataCommand,
)


class Command(BaseExportEntityDataCommand):
    """
    Команда для выгрузки данных сущностей РВД.
    """

    # flake8: noqa: A003
    help = 'Команда для выгрузки данных сущностей РВД за указанных период.'



from django.conf import (
    settings,
)

from edu_rdm_integration.adapters.runners import (
    WebEduRunner,
)
from edu_rdm_integration.consts import (
    LOGS_DELIMITER,
)
from educommon import (
    logger,
)
from educommon.utils.seqtools import (
    make_chunks,
)


class BaseCollectingCalculatingDataRunner(WebEduRunner):
    """
    Базовый класс ранеров функций сбора расчетных данных для интеграции с "Региональная витрина данных".
    """

    def _populate_queue_by_runnable_classes(self, logs, *args, **kwargs):
        """
        Заполнение очереди запускаемыми объектами.
        """
        raw_logs_chunks = make_chunks(
            iterable=logs,
            size=settings.RDM_COLLECT_CHUNK_SIZE,
            is_list=True,
        )

        for chunk_index, raw_logs in enumerate(raw_logs_chunks, start=1):
            runnable_classes = self._prepare_runnable_classes()

            for runnable_class in runnable_classes:
                logger.info(
                    f'{LOGS_DELIMITER*2}enqueue {runnable_class.__name__} with logs chunk {chunk_index} with '
                    f'{len(raw_logs)} records..'
                )

                runnable = runnable_class(raw_logs=raw_logs, *args, **kwargs)

                self.enqueue(runnable=runnable, *args, **kwargs)



from abc import (
    ABCMeta,
)
from typing import (
    Type,
)

from edu_rdm_integration.adapters.helpers import (
    WebEduFunctionHelper,
)
from edu_rdm_integration.adapters.results import (
    WebEduFunctionResult,
)
from edu_rdm_integration.adapters.validators import (
    WebEduFunctionValidator,
)
from function_tools.functions import (
    BaseFunction,
    LazySavingPredefinedQueueFunction,
    LazySavingPredefinedQueueGlobalHelperFunction,
)


class WebEduFunction(BaseFunction, metaclass=ABCMeta):
    """
    Базовый класс для создания функций продуктов Образования.
    """


class WebEduLazySavingPredefinedQueueFunction(LazySavingPredefinedQueueFunction, metaclass=ABCMeta):
    """
    Базовый класс для создания функций с отложенным сохранением объектов
    моделей с предустановленной очередью продуктов Образования.
    """

    def _prepare_helper_class(self) -> Type[WebEduFunctionHelper]:
        """
        Возвращает класс помощника функции.
        """
        return WebEduFunctionHelper

    def _prepare_validator_class(self) -> Type[WebEduFunctionValidator]:
        """
        Возвращает класс валидатора функции.
        """
        return WebEduFunctionValidator

    def _prepare_result_class(self) -> Type[WebEduFunctionResult]:
        """
        Возвращает класс результата функции.
        """
        return WebEduFunctionResult


class WebEduLazySavingPredefinedQueueGlobalHelperFunction(
    LazySavingPredefinedQueueGlobalHelperFunction,
    metaclass=ABCMeta,
):
    """
    Базовый класс для создания функций с отложенным сохранением,
    предустановленной очередью на сохранение и глобальным помощником продуктов Образования.
    """


from datetime import (
    date,
    datetime,
    time,
)
from typing import (
    TYPE_CHECKING,
    List,
    Optional,
    Tuple,
)

from m3_db_utils.models import (
    ModelEnumValue,
)


if TYPE_CHECKING:
    from educommon.audit_log.models import (
        AuditLog,
    )


class BaseEduLogGenerator:
    """
    Базовый класс генератора логов для указанной модели РВД за определенный период времени.

    Для каждой модели РВД есть модели в продукте, создание экземпляров которых, сигнализирует о необходимости сбора
    и выгрузки данных в РВД. Модели можно найти в
    edu_rdm_integration/mapping.py MODEL_FIELDS_LOG_FILTER, принадлежность к конкретной
    модели РВД необходимо определять в функциях.
    """

    def generate(
        self,
        model: ModelEnumValue,
        logs_period_started_at: datetime = datetime.combine(date.today(), time.min),
        logs_period_ended_at: datetime = datetime.combine(date.today(), time.max),
        school_ids: Optional[Tuple[int]] = (),
    ) -> List['AuditLog']:
        """
        Возвращает список сгенерированных экземпляров модели AuditLog.

        Формирование логов производится для указанной модели РВД за указанный период времени.

        Args:
            model: значение модели РВД из модели-перечисления;
            logs_period_started_at: начало периода формирования логов;
            logs_period_ended_at: конец периода формирования логов;
            school_ids: список идентификаторов школ.
        """
        generate_logs_method = getattr(self, f'_generate_{model.key.lower()}_logs')

        logs = generate_logs_method(
            logs_period_started_at=logs_period_started_at,
            logs_period_ended_at=logs_period_ended_at,
            school_ids=school_ids,
        )

        return logs



from datetime import (
    date,
    datetime,
)
from itertools import (
    chain,
)
from typing import (
    List,
    Type,
)

from django.conf import (
    settings,
)

from edu_rdm_integration.adapters.helpers import (
    WebEduFunctionHelper,
    WebEduRunnerHelper,
)
from edu_rdm_integration.consts import (
    DATE_FORMAT,
    EXPORT_DATETIME_FORMAT,
    HASH_ALGORITHM,
)
from edu_rdm_integration.export_data.base.caches import (
    BaseExportDataFunctionCacheStorage,
    BaseExportDataRunnerCacheStorage,
)
from educommon.utils.crypto import (
    HashData,
)


class BaseExportDataRunnerHelper(WebEduRunnerHelper):
    """
    Базовый класс помощников ранеров функций выгрузки данных для интеграции с "Региональная витрина данных".
    """

    def _prepare_cache_class(self) -> Type[BaseExportDataRunnerCacheStorage]:
        """
        Возвращает класс кеша помощника ранера.
        """
        return BaseExportDataRunnerCacheStorage


class BaseExportDataFunctionHelper(WebEduFunctionHelper):
    """
    Базовый класс помощников функций выгрузки данных для интеграции с "Региональная витрина данных".
    """

    cryptographer = HashData(hash_algorithm=HASH_ALGORITHM)

    @classmethod
    def prepare_record(
        cls,
        entity_instance,
        ordered_fields,
        primary_key_fields,
        foreign_key_fields,
        hashable_fields
    ) -> List[str]:
        """
        Формирование списка строковых значений полей.
        """
        field_values = []
        key_fields = set(chain(primary_key_fields, foreign_key_fields))

        for field in ordered_fields:
            field_value = getattr(entity_instance, field)

            if isinstance(field_value, datetime):
                # Дату/время передаём в формате: YYYY-MM-DD hh:mm:ss
                field_value = field_value.strftime(EXPORT_DATETIME_FORMAT)
            elif isinstance(field_value, date):
                field_value = field_value.strftime(DATE_FORMAT)
            else:
                field_value = str(field_value if field_value is not None else '')

            if field_value and field in key_fields:
                field_value = f'{settings.RDM_EXPORT_ENTITY_ID_PREFIX}-{field_value}'

            if field_value and field in hashable_fields:
                field_value = cls.cryptographer.get_hash(field_value)

            # Экранирование двойных кавычек
            field_value = field_value.replace('"', '""')

            field_values.append(field_value)

        return [f'"{v}"' for v in field_values]

    def _prepare_cache_class(self) -> Type[BaseExportDataFunctionCacheStorage]:
        """
        Возвращает класс кеша помощника функции.
        """
        return BaseExportDataFunctionCacheStorage



from abc import (
    ABCMeta,
)
from typing import (
    Type,
)

from edu_rdm_integration.adapters.runners import (
    WebEduRunner,
)
from function_tools.managers import (
    RunnerManager,
)


class WebEduRunnerManager(RunnerManager, metaclass=ABCMeta):
    """
    Базовый класс менеджеров пускателей функций продуктов Образования.
    """

    @classmethod
    def _prepare_runner_class(cls) -> Type[WebEduRunner]:
        """
        Возвращает класс ранера.
        """
        return WebEduRunner




from collections import (
    defaultdict,
)
from typing import (
    Any,
    Dict,
    List,
    NamedTuple,
    Union,
)

from edu_rdm_integration.adapters.caches import (
    WebEduFunctionCacheStorage,
    WebEduRunnerCacheStorage,
)
from edu_rdm_integration.mapping import (
    MODEL_FIELDS_LOG_FILTER,
)
from educommon.audit_log.utils import (
    get_model_by_table,
)
from educommon.integration_entities.consts import (
    LOG_OPERATION_MAP,
)
from educommon.integration_entities.enums import (
    EntityLogOperation,
)
from educommon.utils.conversion import (
    int_or_none,
)


class LogChange(NamedTuple):
    """Операция и значения измененных полей из лога."""

    operation: EntityLogOperation
    fields: Dict[str, Any]

    @property
    def is_create(self) -> bool:
        """Лог создания."""
        return self.operation == EntityLogOperation.CREATE

    @property
    def is_update(self) -> bool:
        """Лог изменения."""
        return self.operation == EntityLogOperation.UPDATE

    @property
    def is_delete(self) -> bool:
        """Лог удаления."""
        return self.operation == EntityLogOperation.DELETE


class BaseCollectingExportedDataRunnerCacheStorage(WebEduRunnerCacheStorage):
    """
    Базовый кеш помощников ранеров функций сбора данных для интеграции с "Региональная витрина данных".
    """


class BaseCollectingExportedDataFunctionCacheStorage(WebEduFunctionCacheStorage):
    """
    Базовый кеш помощников функций сбора данных для интеграции с "Региональная витрина данных".
    """

    def __init__(self, raw_logs, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Необработанные логи как есть
        self.raw_logs = raw_logs

        # TODO Перенести в базовый класс (https://jira.bars.group/browse/EDUSCHL-19991)
        # Необходимость объединения логов относящихся к одному объекту
        self.is_merge_logs = True

        # Подготовленные логи в виде:
        # {
        #     'person.person': {
        #         48939: [(1, {'surname': 'Иванов', 'snils': '157-283-394 92'...}), (2, {...})],
        #         44281: [(2, {...}), (2, {...}), ...],
        #         12600: [(3, {...})],
        #     },
        #     'schoolchild.schoolchild': {
        #         ...
        #     },
        # }
        self.logs: Dict[str, Dict[int, List[LogChange]]] = defaultdict(lambda: defaultdict(list))

        self._prepare_logs()

    @staticmethod
    def _filter_log(model: str, log_change: LogChange) -> bool:
        """
        Производится проверка изменений на отслеживаемые поля.
        """
        is_filtered = False

        if model in MODEL_FIELDS_LOG_FILTER[log_change.operation]:
            filter_fields = MODEL_FIELDS_LOG_FILTER[log_change.operation][model]
            if filter_fields:
                # Если заданы конкретные поля, которые должны отслеживать
                for field in log_change.fields:
                    if field in filter_fields:
                        # Достаточно, чтобы хотя бы одно поле попало под фильтр
                        is_filtered = True
                        break
            else:
                # Модель отслеживается, но перечень фильтруемых полей не задан,
                # значит фильтруем все поля модели
                is_filtered = True

        return is_filtered

    @staticmethod
    def _add_id_to_set(ids: set, _id: Union[int, str, None]) -> None:
        """Добавление значения id во множество ids."""
        _id = int_or_none(_id)
        if _id:
            ids.add(_id)

    def _reformat_logs(self):
        """
        Производится преобразование логов к удобному для работы виду.

        Предполагается вложенные словари. На первом уровне ключом будет название модели, на втором идентификатор записи.
        """
        for log in self.raw_logs:
            model = get_model_by_table(log.table)._meta.label
            operation = LOG_OPERATION_MAP[log.operation]

            if operation in (EntityLogOperation.CREATE, EntityLogOperation.DELETE):
                fields = log.data
            elif operation == EntityLogOperation.UPDATE:
                fields = log.changes
            else:
                fields = {}

            log_change = LogChange(
                operation=operation,
                fields=fields,
            )

            if not self._filter_log(model, log_change):
                # Если модель не отслеживается, то запись лога не сохраняем
                continue

            if log_change.operation == EntityLogOperation.DELETE:
                self.logs[model][log.object_id] = [log_change, ]
            else:
                self.logs[model][log.object_id].append(log_change)

    def _prepare_logs(self):
        """
        Подготовка логов для дальнейшей работы.
        """
        self._reformat_logs()



# Generated by Django 2.2.28 on 2023-07-13 17:32

import django.db.models.deletion
from django.db import (
    migrations,
    models,
)

import edu_rdm_integration.utils
import educommon.django.db.mixins
import educommon.utils.date


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('uploader_client', '0001_initial'),
        ('function_tools', '0005_auto_20220724_0050'),
    ]

    operations = [
        migrations.CreateModel(
            name='CollectingDataStageStatus',
            fields=[
                ('title', models.TextField(verbose_name='расшифровка значения')),
                ('key', models.CharField(db_index=True, max_length=512, primary_key=True, serialize=False, verbose_name='ключ')),
                ('order_number', models.PositiveIntegerField(default=100000, verbose_name='Порядковый номер')),
            ],
            options={
                'verbose_name': 'Модель-перечисление статусов этапа сбора данных',
                'verbose_name_plural': 'Модели-перечисления статусов этапов сбора данных',
                'db_table': 'rdm_collecting_data_stage_status',
            },
        ),
        migrations.CreateModel(
            name='CollectingDataSubStageStatus',
            fields=[
                ('title', models.TextField(verbose_name='расшифровка значения')),
                ('key', models.CharField(db_index=True, max_length=512, primary_key=True, serialize=False, verbose_name='ключ')),
                ('order_number', models.PositiveIntegerField(default=100000, verbose_name='Порядковый номер')),
            ],
            options={
                'verbose_name': 'Модель-перечисление статусов подэтапа сбора данных',
                'verbose_name_plural': 'Модели-перечисления статусов подэтапов сбора данных',
                'db_table': 'rdm_collecting_data_sub_stage_status',
            },
        ),
        migrations.CreateModel(
            name='CollectingExportedDataStage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('logs_period_started_at', models.DateTimeField(db_index=True, default=educommon.utils.date.get_today_min_datetime, verbose_name='Левая граница периода обрабатываемых логов')),
                ('logs_period_ended_at', models.DateTimeField(db_index=True, default=educommon.utils.date.get_today_max_datetime, verbose_name='Правая граница периода обрабатываемых логов')),
                ('started_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Время начала сбора данных')),
                ('ended_at', models.DateTimeField(blank=True, db_index=True, null=True, verbose_name='Время завершения сбора данных')),
                ('manager', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='function_tools.Entity', verbose_name='Менеджер ранера Функции')),
                ('status', models.ForeignKey(default='CREATED', on_delete=django.db.models.deletion.PROTECT, to='edu_rdm_integration.CollectingDataStageStatus', verbose_name='Статус')),
            ],
            options={
                'verbose_name': 'Этап формирования данных для выгрузки',
                'verbose_name_plural': 'Этапы формирования данных для выгрузки',
                'db_table': 'rdm_collecting_exported_data_stage',
            },
            bases=(educommon.django.db.mixins.ReprStrPreModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ExportingDataStage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('period_started_at', models.DateTimeField(db_index=True, verbose_name='Левая граница периода выборки данных для выгрузки')),
                ('period_ended_at', models.DateTimeField(db_index=True, verbose_name='Правая граница периода выборки данных для выгрузки')),
                ('started_at', models.DateTimeField(auto_now_add=True, verbose_name='Время начала выгрузки данных')),
                ('ended_at', models.DateTimeField(blank=True, null=True, verbose_name='Время завершения выгрузки данных')),
                ('manager', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='function_tools.Entity', verbose_name='Менеджер ранера Функции')),
            ],
            options={
                'verbose_name': 'Этап формирования данных для выгрузки',
                'verbose_name_plural': 'Этапы формирования данных для выгрузки',
                'db_table': 'rdm_exporting_data_stage',
            },
            bases=(educommon.django.db.mixins.ReprStrPreModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ExportingDataStageStatus',
            fields=[
                ('title', models.TextField(verbose_name='расшифровка значения')),
                ('key', models.CharField(db_index=True, max_length=512, primary_key=True, serialize=False, verbose_name='ключ')),
                ('order_number', models.PositiveIntegerField(default=100000, verbose_name='Порядковый номер')),
            ],
            options={
                'verbose_name': 'Модель-перечисление статусов этапа выгрузки данных',
                'verbose_name_plural': 'Модели-перечисления статусов этапов выгрузки данных',
                'db_table': 'rdm_exporting_data_stage_status',
            },
        ),
        migrations.CreateModel(
            name='ExportingDataSubStage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('started_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Время начала сбора данных')),
                ('ended_at', models.DateTimeField(blank=True, db_index=True, null=True, verbose_name='Время завершения сбора данных')),
                ('function', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='function_tools.Entity', verbose_name='Функция')),
                ('stage', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='edu_rdm_integration.ExportingDataStage', verbose_name='Этап выгрузки данных')),
            ],
            options={
                'verbose_name': 'Стадия выгрузки данных',
                'verbose_name_plural': 'Стадии выгрузки данных',
                'db_table': 'rdm_exporting_data_sub_stage',
            },
            bases=(educommon.django.db.mixins.ReprStrPreModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ExportingDataSubStageAttachment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('attachment', models.FileField(blank=True, max_length=512, null=True, upload_to=edu_rdm_integration.utils.get_exporting_data_stage_attachment_path, verbose_name='Сгенерированный файл')),
                ('operation', models.SmallIntegerField(choices=[(1, 'Создание'), (2, 'Изменение'), (3, 'Удаление')], verbose_name='Действие')),
                ('created', models.DateTimeField(auto_now_add=True, null=True, verbose_name='Дата создания')),
                ('modified', models.DateTimeField(auto_now=True, null=True, verbose_name='Дата изменения')),
                ('exporting_data_sub_stage', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='edu_rdm_integration.ExportingDataSubStage', verbose_name='Подэтап выгрузки данных')),
            ],
            options={
                'verbose_name': 'Сгенерированный файл для дальнейшей выгрузки в "Региональная витрина данных"',
                'verbose_name_plural': 'Сгенерированные файлы для дальнейшей выгрузки в "Региональная витрина данных"',
                'db_table': 'rdm_exporting_data_sub_stage_attachment',
            },
            bases=(educommon.django.db.mixins.ReprStrPreModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ExportingDataSubStageStatus',
            fields=[
                ('title', models.TextField(verbose_name='расшифровка значения')),
                ('key', models.CharField(db_index=True, max_length=512, primary_key=True, serialize=False, verbose_name='ключ')),
                ('order_number', models.PositiveIntegerField(default=100000, verbose_name='Порядковый номер')),
            ],
            options={
                'verbose_name': 'Модель-перечисление статусов подэтапа выгрузки данных',
                'verbose_name_plural': 'Модели-перечисления статусов подэтапов выгрузки данных',
                'db_table': 'rdm_exporting_data_sub_stage_status',
            },
        ),
        migrations.CreateModel(
            name='ExportingDataSubStageUploaderClientLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('request_id', models.CharField(blank=True, max_length=100, verbose_name='Id запроса загрузки в витрину')),
                ('is_emulation', models.BooleanField(default=False, verbose_name='Включен режим эмуляции')),
                ('file_upload_status', models.SmallIntegerField(blank=True, choices=[(1, 'В процессе загрузки в витрину'), (2, 'Загрузка в витрину закончена'), (3, 'Ошибка загрузки в витрину')], null=True, verbose_name='Общий статус загрузки файла в витрину')),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True, null=True, verbose_name='Дата создания')),
                ('modified', models.DateTimeField(auto_now=True, db_index=True, null=True, verbose_name='Дата изменения')),
                ('attachment', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='edu_rdm_integration.ExportingDataSubStageAttachment', verbose_name='Прикрепленный файл')),
                ('entry', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='uploader_client_log', to='uploader_client.Entry', verbose_name='Лог запроса и ответа')),
                ('sub_stage', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='edu_rdm_integration.ExportingDataSubStage', verbose_name='Подэтап выгрузки данных')),
            ],
            options={
                'verbose_name': 'Лог запроса подэтапа выгрузки данных',
                'verbose_name_plural': 'Лог запроса подэтапа выгрузки данных',
                'db_table': 'rdm_exporting_data_sub_stage_uploader_client_log',
            },
            bases=(educommon.django.db.mixins.ReprStrPreModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='RegionalDataMartEntityEnum',
            fields=[
                ('title', models.TextField(verbose_name='расшифровка значения')),
                ('key', models.CharField(db_index=True, max_length=512, primary_key=True, serialize=False, verbose_name='ключ')),
                ('order_number', models.PositiveIntegerField(default=100000, verbose_name='Порядковый номер')),
            ],
            options={
                'verbose_name': 'Модель-перечисление сущностей "Региональной витрины данных"',
                'verbose_name_plural': 'Модели-перечисления сущностей "Региональной витрины данных"',
                'db_table': 'rdm_entity',
            },
        ),
        migrations.CreateModel(
            name='RegionalDataMartModelEnum',
            fields=[
                ('title', models.TextField(verbose_name='расшифровка значения')),
                ('key', models.CharField(db_index=True, max_length=512, primary_key=True, serialize=False, verbose_name='ключ')),
                ('order_number', models.PositiveIntegerField(default=100000, verbose_name='Порядковый номер')),
            ],
            options={
                'verbose_name': 'Модель-перечисление моделей "Региональной витрины данных"',
                'verbose_name_plural': 'Модели-перечисления моделей "Региональной витрины данных"',
                'db_table': 'rdm_model',
            },
        ),
        migrations.CreateModel(
            name='UploadStatus',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.SmallIntegerField(blank=True, null=True, verbose_name='Код статуса загрузки в витрину')),
                ('description', models.CharField(max_length=100, verbose_name='Описание статуса загрузки в витрину')),
            ],
            options={
                'verbose_name': 'Статус загрузки данных в витрину',
                'verbose_name_plural': 'Статусы загрузки данных в витрину',
                'db_table': 'rdm_upload_status',
            },
            bases=(educommon.django.db.mixins.ReprStrPreModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='UploadStatusRequestLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('entry', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='upload_status_request_log', to='uploader_client.Entry', verbose_name='Cвязь запроса статуса с запросом в витрину')),
                ('status', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='edu_rdm_integration.UploadStatus', verbose_name='Cвязь с таблицей возможных статусов')),
                ('upload', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='edu_rdm_integration.ExportingDataSubStageUploaderClientLog', verbose_name='Cвязь запроса статуса с загрузкой файла в витрину')),
            ],
            options={
                'verbose_name': 'Лог запроса подэтапа выгрузки данных',
                'verbose_name_plural': 'Логи запроса подэтапа выгрузки данных',
                'db_table': 'rdm_upload_status_request_log',
            },
            bases=(educommon.django.db.mixins.ReprStrPreModelMixin, models.Model),
        ),
        migrations.AddField(
            model_name='exportingdatasubstage',
            name='status',
            field=models.ForeignKey(default='CREATED', on_delete=django.db.models.deletion.PROTECT, to='edu_rdm_integration.ExportingDataSubStageStatus', verbose_name='Статус'),
        ),
        migrations.AddField(
            model_name='exportingdatastage',
            name='status',
            field=models.ForeignKey(default='CREATED', on_delete=django.db.models.deletion.PROTECT, to='edu_rdm_integration.ExportingDataStageStatus', verbose_name='Статус'),
        ),
        migrations.CreateModel(
            name='CollectingExportedDataSubStage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('started_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Время начала сбора данных')),
                ('ended_at', models.DateTimeField(blank=True, db_index=True, null=True, verbose_name='Время завершения сбора данных')),
                ('function', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='function_tools.Entity', verbose_name='Функция')),
                ('previous', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='edu_rdm_integration.CollectingExportedDataSubStage', verbose_name='Предыдущий сбор данных')),
                ('stage', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='edu_rdm_integration.CollectingExportedDataStage', verbose_name='Этап подготовки данных для экспорта')),
                ('status', models.ForeignKey(default='CREATED', on_delete=django.db.models.deletion.PROTECT, to='edu_rdm_integration.CollectingDataSubStageStatus', verbose_name='Статус')),
            ],
            options={
                'verbose_name': 'Подэтап формирования данных для выгрузки',
                'verbose_name_plural': 'Подэтапы формирования данных для выгрузки',
                'db_table': 'rdm_collecting_exported_data_sub_stage',
            },
            bases=(educommon.django.db.mixins.ReprStrPreModelMixin, models.Model),
        ),
    ]



from abc import (
    abstractmethod,
)
from datetime import (
    date,
    datetime,
    time,
    timedelta,
)
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Type,
)

from django.conf import (
    settings,
)
from django.core.management import (
    CommandParser,
)
from django.core.management.base import (
    BaseCommand,
)

from edu_rdm_integration.consts import (
    DATETIME_FORMAT,
    REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA,
    REGIONAL_DATA_MART_INTEGRATION_EXPORTING_DATA,
)
from edu_rdm_integration.management.generators import (
    BaseEduLogGenerator,
)
from edu_rdm_integration.models import (
    RegionalDataMartEntityEnum,
    RegionalDataMartModelEnum,
)
from edu_rdm_integration.storages import (
    RegionalDataMartEntityStorage,
)
from educommon import (
    logger,
)
from educommon.audit_log.models import (
    AuditLog,
)
from function_tools.managers import (
    RunnerManager,
)


if TYPE_CHECKING:
    from django.core.management.base import (
        CommandParser,
    )


class BaseCollectModelDataCommand(BaseCommand):
    """
    Базовая команда для выполнения сбора данных моделей РВД.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Классы менеджеров Функций, которые должны быть запущены для сбора данных моделей РВД
        self._collecting_data_managers: Dict[str, Type[RunnerManager]] = dict()

        # Результаты работы Функций сбора данных моделей РВД
        self._collecting_data_results = []

    def add_arguments(self, parser: 'CommandParser'):
        """
        Добавление параметров.
        """
        models = ', '.join([
            f'{key} - {value.title}'
            for key, value in RegionalDataMartModelEnum.get_enum_data().items()
        ])
        models_help_text = (
            f'Значением параметра является перечисление моделей РВД, для которых должен быть произведен сбор данных. '
            f'Перечисление моделей:\n{models}. Если модели не указываются, то сбор данных производится для всех '
            f'моделей. Модели перечисляются через запятую без пробелов.'
        )
        parser.add_argument(
            '--models',
            action='store',
            dest='models',
            type=(
                lambda e: tuple(
                    map(
                        lambda e_str: RegionalDataMartModelEnum.get_model_enum_value(e_str),
                        map(str.strip, e.split(','))
                    )
                )
            ),
            default=tuple(value for value in RegionalDataMartModelEnum.get_enum_data().values()),
            help=models_help_text,
        )

        parser.add_argument(
            '--logs_period_started_at',
            action='store',
            dest='logs_period_started_at',
            type=lambda started_at: datetime.strptime(started_at, DATETIME_FORMAT),
            default=datetime.combine(date.today(), time.min),
            help=(
                'Дата и время начала периода обрабатываемых логов. Значение предоставляется в формате '
                '"дд.мм.гггг чч:мм:сс". По умолчанию, сегодняшний день, время 00:00:00.'
            ),
        )

        parser.add_argument(
            '--logs_period_ended_at',
            action='store',
            dest='logs_period_ended_at',
            type=lambda ended_at: datetime.strptime(ended_at, DATETIME_FORMAT),
            default=datetime.combine(date.today(), time.max),
            help=(
                'Дата и время конца периода обрабатываемых логов. Значение предоставляется в формате '
                '"дд.мм.гггг чч:мм:сс". По умолчанию, сегодняшний день, время 23:59:59.'
            ),
        )

    def _find_collecting_models_data_managers(self, *args, **kwargs):
        """
        Поиск менеджеров Функций, которые должны быть запущены для сбора данных моделей РВД.
        """
        logger.info('collecting_models_data_managers..')

        entity_storage = RegionalDataMartEntityStorage()
        entity_storage.prepare()

        collecting_models_data_managers_map = entity_storage.prepare_entities_manager_map(
            tags={REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA},
        )

        for model_enum in kwargs.get('models'):
            manager_class = collecting_models_data_managers_map.get(model_enum.key)

            if manager_class:
                self._collecting_data_managers[model_enum.key] = manager_class

        logger.info('collecting models data managers finished')

    def _collect_models_data(self, *args, logs: Optional[Dict[str, List[AuditLog]]] = None, **kwargs):
        """
        Запуск Функций по формированию данных моделей РВД из логов.
        """
        logger.info('collect models data..')

        for model_key, manager_class in self._collecting_data_managers.items():
            model_logs = logs.get(model_key) if logs else None
            manager = manager_class(*args, logs=model_logs, **kwargs)
            manager.run()

            self._collecting_data_results.append(manager.result)

        logger.info('collecting entities data finished.')

    def handle(self, *args, **kwargs):
        """
        Выполнение действий команды.
        """
        logger.info(
            f'start collecting data of models - {", ".join([model.key for model in kwargs["models"]])}..'
        )

        self._find_collecting_models_data_managers(*args, **kwargs)
        self._collect_models_data(*args, **kwargs)

        logger.info('collecting models data finished.')


class BaseExportEntityDataCommand(BaseCommand):
    """
    Базовая команда для выполнения выгрузки данных сущностей РВД за указанный период.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Классы менеджеров Функций, которые должны быть запущены для выгрузки данных
        self._exporting_data_managers: Set[Type[RunnerManager]] = set()

        # Результаты работы Функций выгрузки данных
        self._exporting_data_results = []

        self._configure_agent_client()

    def add_arguments(self, parser: 'CommandParser'):
        """
        Добавление параметров.
        """
        entities = ', '.join([
            f'{key} - {value.title}'
            for key, value in RegionalDataMartEntityEnum.get_enum_data().items()
        ])
        entities_help_text = (
            f'Значением параметра является перечисление сущностей РВД, для которых должена быть произведена выгрузка '
            f'данных. Перечисление сущностей:\n{entities}. Если сущности не указываются, то выгрузка данных '
            f'производится для всех сущностей. Сущности перечисляются через запятую без пробелов.'
        )
        parser.add_argument(
            '--entities',
            action='store',
            dest='entities',
            type=(
                lambda e: tuple(
                    map(
                        lambda e_str: RegionalDataMartEntityEnum.get_model_enum_value(e_str),
                        map(str.strip, e.split(','))
                    )
                )
            ),
            default=tuple(value for value in RegionalDataMartEntityEnum.get_enum_data().values()),
            help=entities_help_text,
        )

        parser.add_argument(
            '--period_started_at',
            action='store',
            dest='period_started_at',
            type=lambda started_at: datetime.strptime(started_at, DATETIME_FORMAT),
            default=datetime.combine(date.today(), time.min),
            help=(
                'Дата и время начала периода сбора записей моделей РВД. Значение предоставляется в формате '
                '"дд.мм.гггг чч:мм:сс". По умолчанию, сегодняшний день, время 00:00:00.'
            ),
        )

        parser.add_argument(
            '--period_ended_at',
            action='store',
            dest='period_ended_at',
            type=(
                lambda ended_at: datetime.strptime(ended_at, DATETIME_FORMAT).replace(microsecond=time.max.microsecond)
            ),
            default=datetime.combine(date.today(), time.max),
            help=(
                'Дата и время конца периода сбора записей моделей РВД. Значение предоставляется в формате '
                '"дд.мм.гггг чч:мм:сс". По умолчанию, сегодняшний день, время 23:59:59.'
            ),
        )
        parser.add_argument(
            '--task_id',
            action='store',
            dest='task_id',
            type=str,
            default=None,
            help='task_id для поиска асинхронной задачи',
        )

    def _configure_agent_client(self):
        """
        Конфигурирование клиента загрузчика данных в Витрину.

        #TODO Вынужденная мера, т.к. при запуске команды не производится проверка готовности конфигов приложений.
          # Нужно переработать механизм конфигурирования клиента загрузчика.
        """
        import uploader_client
        from uploader_client.contrib.rdm.interfaces.configurations import (
            RegionalDataMartUploaderConfig,
        )
        if settings.RDM_UPLOADER_CLIENT_ENABLE_REQUEST_EMULATION:
            uploader_client.set_config(
                RegionalDataMartUploaderConfig(
                    interface='uploader_client.contrib.rdm.interfaces.rest.OpenAPIInterfaceEmulation',
                    url=settings.RDM_UPLOADER_CLIENT_URL,
                    datamart_name=settings.RDM_UPLOADER_CLIENT_DATAMART_NAME,
                    timeout=1,
                    request_retries=1,
                )
            )
        else:
            uploader_client.set_config(
                RegionalDataMartUploaderConfig(
                    url=settings.RDM_UPLOADER_CLIENT_URL,
                    datamart_name=settings.RDM_UPLOADER_CLIENT_DATAMART_NAME,
                    timeout=settings.RDM_UPLOADER_CLIENT_REQUEST_TIMEOUT,
                    request_retries=settings.RDM_UPLOADER_CLIENT_REQUEST_RETRIES,
                )
            )

    def _find_exporting_entities_data_managers(self, *args, **kwargs):
        """
        Поиск менеджеров Функций выгрузки данных по сущностям РВД.
        """
        logger.info('find exporting entities data manager..')

        entity_storage = RegionalDataMartEntityStorage()
        entity_storage.prepare()

        exporting_entities_data_managers_map = entity_storage.prepare_entities_manager_map(
            tags={REGIONAL_DATA_MART_INTEGRATION_EXPORTING_DATA},
        )

        for entity in kwargs.get('entities'):
            manager_class = exporting_entities_data_managers_map.get(entity.key)

            if manager_class:
                self._exporting_data_managers.add(manager_class)

        logger.info('finding exporting entities data manager finished.')

    def _export_entities_data(self, *args, **kwargs):
        """
        Выгрузка данных по указанным сущностям.
        """
        logger.info('start exporting entities data..')

        for manager_class in self._exporting_data_managers:
            manager = manager_class(*args, is_only_main_model=True, **kwargs)
            manager.run()

            self._exporting_data_results.append(manager.result)

        logger.info('exporting entities data finished.')

    def handle(self, *args, **kwargs):
        """
        Выполнение действий команды.
        """
        logger.info(
            f'start exporting data of entities - {", ".join([entity.key for entity in kwargs["entities"]])}..'
        )

        self._find_exporting_entities_data_managers(*args, **kwargs)
        self._export_entities_data(*args, **kwargs)

        logger.info('exporting entities data finished.')


class BaseCollectModelsDataByGeneratingLogsCommand(BaseCollectModelDataCommand):
    """
    Команда сбора данных моделей РВД на основе существующих в БД данных моделей ЭШ.

    Можно регулировать, для каких моделей должен быть произведен сбор данных, и период, за который должны
    быть собраны логи. Логи формируются в процессе выполнения команды при помощи генератора логов LogGenerator для
    указанной модели.
    """

    # flake8: noqa: A003
    help = 'Команда сбора данных моделей РВД на основе существующих в БД данных моделей продукта'

    def __init__(self, *args, **kwargs):
        """Инициализация команды."""
        super().__init__(*args, **kwargs)

        self.log_generator = self.prepare_log_generator()

    @abstractmethod
    def prepare_log_generator(self) -> 'BaseEduLogGenerator':
        """Возвращает генератор логов."""

    def add_arguments(self, parser: 'CommandParser'):
        """
        Добавление параметров.
        """
        super().add_arguments(parser=parser)

        parser.add_argument(
            '--logs_sub_period_days',
            action='store',
            dest='logs_sub_period_days',
            type=int,
            default=0,
            help=(
                'Размер подпериодов, на которые будет разбит основной период, в днях. По умолчанию, '
                '0 - разбиение на подпериоды отключено.'
            ),
        )

        parser.add_argument(
            '--school_ids',
            action='store',
            dest='school_ids',
            type=lambda v: tuple(map(int, v.split(','))),
            default=(),
            help='Школы, для которых производится выгрузка.',
        )

    def _generate_logs_by_subperiod(self, *args, **kwargs):
        """
        Генерация логов с учетом подпериодов.
        """
        logs_period_started_at = kwargs.get('logs_period_started_at')
        logs_period_ended_at = kwargs.get('logs_period_ended_at')
        school_ids = kwargs.get('school_ids')

        temp_logs_period_started_at = logs_period_started_at
        temp_logs_period_ended_at = logs_period_started_at + timedelta(days=kwargs.get('logs_sub_period_days'))

        if temp_logs_period_ended_at > logs_period_ended_at:
            temp_logs_period_ended_at = logs_period_ended_at

        temp_logs: Dict[str, List[AuditLog]] = dict()

        while temp_logs_period_started_at < temp_logs_period_ended_at <= logs_period_ended_at:
            for model in kwargs.get('models'):
                logs = self.log_generator.generate(
                    model=model,
                    logs_period_started_at=temp_logs_period_started_at,
                    logs_period_ended_at=temp_logs_period_ended_at,
                    school_ids=school_ids,
                )

                temp_logs[model.key] = logs

            yield temp_logs, temp_logs_period_started_at, temp_logs_period_ended_at

            temp_logs_period_started_at = temp_logs_period_ended_at
            temp_logs_period_ended_at += timedelta(days=kwargs.get('logs_sub_period_days'))

            if temp_logs_period_ended_at > logs_period_ended_at:
                temp_logs_period_ended_at = logs_period_ended_at

            temp_logs.clear()

    def _generate_logs_for_all_period(self, *args, **kwargs):
        """
        Генерация логов за весь период.
        """
        logs_period_started_at = kwargs.get('logs_period_started_at')
        logs_period_ended_at = kwargs.get('logs_period_ended_at')
        school_ids = kwargs.get('school_ids')

        temp_logs: Dict[str, List[AuditLog]] = dict()

        for model in kwargs.get('models'):
            logs = self.log_generator.generate(
                model=model,
                logs_period_started_at=logs_period_started_at,
                logs_period_ended_at=logs_period_ended_at,
                school_ids=school_ids,
            )

            temp_logs[model.key] = logs

        return [(temp_logs, logs_period_started_at, logs_period_ended_at)]

    def _generate_logs(self, *args, **kwargs) -> List[Tuple[Dict[str, List[AuditLog]], datetime, datetime]]:
        """
        Генерация логов.

        Осуществляет генерацию логов по уже существующим записям в базе данных. В качестве параметров указываются
        начало и конец периода сбора логов, размер подпериодов, на которые должен быть разбит основной период.
        Генерация логов производится только для указанных моделей.
        """
        if kwargs.get('logs_sub_period_days'):
            logs = self._generate_logs_by_subperiod(*args, **kwargs)
        else:
            logs = self._generate_logs_for_all_period(*args, **kwargs)

        return logs

    def handle(self, *args, **kwargs):
        """
        Запуск сбора данных указанных моделей РВД на основе существующих данных моделей ЭШ.
        """
        logger.info(
            f'start collecting data of models - {", ".join([model.key for model in kwargs["models"]])}..'
        )

        self._find_collecting_models_data_managers(*args, **kwargs)

        temp_kwargs = kwargs.copy()

        for logs, logs_period_started_at, logs_period_ended_at in self._generate_logs(*args, **kwargs):
            temp_kwargs['logs_period_started_at'] = logs_period_started_at
            temp_kwargs['logs_period_ended_at'] = logs_period_ended_at

            self._collect_models_data(*args, logs=logs, **temp_kwargs)

        logger.info('collecting models data finished.')



from edu_rdm_integration.management.general import (
    BaseCollectModelDataCommand,
)


class Command(BaseCollectModelDataCommand):
    """
    Команда для сбора данных моделей РВД за указанных период по существующим логам.
    """

    # flake8: noqa: A003
    help = 'Команда для сбора данных моделей РВД за указанных период по существующим логам.'



from abc import (
    ABCMeta,
)
from datetime import (
    date,
    datetime,
    time,
)
from typing import (
    Dict,
    Tuple,
    Type,
)

from edu_rdm_integration.adapters.managers import (
    WebEduRunnerManager,
)
from edu_rdm_integration.consts import (
    LOGS_DELIMITER,
)
from edu_rdm_integration.export_data.base.runners import (
    BaseExportDataRunner,
)
from edu_rdm_integration.models import (
    CollectingDataStageStatus,
    ExportingDataStage,
    ExportingDataStageStatus,
)
from edu_rdm_integration.storages import (
    RegionalDataMartEntityStorage,
)
from educommon import (
    logger,
)
from m3_db_utils.models import (
    ModelEnumValue,
)


class BaseExportDataRunnerManager(WebEduRunnerManager, metaclass=ABCMeta):
    """
    Менеджер ранеров функций выгрузки данных для интеграции с "Региональная витрина данных".
    """

    def __init__(
        self,
        *args,
        is_only_main_model: bool = False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        # Выгрузка данных производится только для основной модели сущности
        self._is_only_main_model = is_only_main_model

        self._period_started_at, self._period_ended_at = self._prepare_period(*args, **kwargs)

        self._stage = ExportingDataStage.objects.create(
            manager_id=self.uuid,
            period_started_at=self._period_started_at,
            period_ended_at=self._period_ended_at,
        )

        logger.info(f'{LOGS_DELIMITER}{repr(self._stage)} created.')

    @classmethod
    def _prepare_runner_class(cls) -> Type[BaseExportDataRunner]:
        """
        Возвращает класс ранера.
        """
        return BaseExportDataRunner

    def _find_exporting_data_stage(self):
        """
        Поиск последнего подэтапа выгрузки данных сущности РВД.
        """
        entity_storage = RegionalDataMartEntityStorage()
        entity_storage.prepare()

        exporting_data_stage = ExportingDataStage.objects.filter(
            manager_id=self.uuid,
            status_id__in=(
                CollectingDataStageStatus.FAILED.key,
                CollectingDataStageStatus.FINISHED.key,
            ),
        ).latest('period_ended_at')

        if exporting_data_stage:
            logger.info(f'{LOGS_DELIMITER}{repr(exporting_data_stage)} sub stages found.')
        else:
            logger.info(f'{LOGS_DELIMITER} sub stages not found.')

        return exporting_data_stage

    def _prepare_period(self, *args, **kwargs) -> Tuple[datetime, datetime]:
        """
        Формирование периода сбора данных моделей РВД.
        """
        period_started_at = kwargs.get('period_started_at')
        period_ended_at = kwargs.get('period_ended_at')

        if not period_started_at:
            last_exporting_data_stage = self._find_exporting_data_stage()

            if last_exporting_data_stage:
                period_started_at = last_exporting_data_stage.period_ended_at
            else:
                period_started_at = datetime.combine(date.today(), time.min)

        if not period_ended_at:
            period_ended_at = datetime.combine(date.today(), time.max)

        return period_started_at, period_ended_at

    def _prepare_model_ids_map(self) -> Dict[ModelEnumValue, Tuple[int]]:
        """
        Осуществляется поиск записей моделей добавленных или обновленных за указанный период времени.

        Т.к. моделей влияющих на сущность может быть множество, то в методе формируется словарь, содержащий в качестве
        ключа название модели, значение - кортеж идентификаторов записей.
        """
        return {}

    def _create_runner(self, *args, **kwargs):
        """
        Производится расширение для осуществления поиска идентификаторов записей моделей РВД для дальнейшей выгрузки.

        model_ids_map пробрасывается в ранер для дальнейшей обработки записей для формирования чанков.
        """
        model_ids_map = self._prepare_model_ids_map()

        super()._create_runner(
            *args,
            model_ids_map=model_ids_map,
            stage=self._stage,
            **kwargs,
        )

    def _before_start_runner(self, *args, **kwargs):
        """
        Точка расширения поведения менеджера ранера перед запуском ранера.
        """
        self._stage.status_id = ExportingDataStageStatus.IN_PROGRESS.key
        self._stage.save()

        logger.info(f'{LOGS_DELIMITER}change status {repr(self._stage)}')

    def _after_start_runner(self, *args, **kwargs):
        """
        Точка расширения поведения менеджера ранера после запуска ранера.
        """
        if self._runner.result.has_not_errors:
            self._stage.status_id = ExportingDataStageStatus.FINISHED.key
        else:
            self._stage.status_id = ExportingDataStageStatus.FAILED.key

        self._stage.save()

        logger.info(f'{LOGS_DELIMITER}change status {repr(self._stage)}')



import os
from datetime import (
    datetime,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
)

from django.conf import (
    settings,
)

from edu_rdm_integration.apps import (
    EduRDMIntegrationConfig,
)


if TYPE_CHECKING:
    from edu_rdm_integration.models import (
        BaseEntityModel,
    )


def get_exporting_data_stage_attachment_path(instance, filename):
    """Возвращает путь до файла-вложения в этап выгрузки данных сущности.

    Args:
        instance: объект ExportingDataStage
        filename: имя загружаемого файла

    Returns:
        Строковое представление пути
    """
    datetime_now = datetime.now()

    return os.path.join(
        settings.UPLOADS,
        EduRDMIntegrationConfig.label,
        datetime_now.strftime('%Y/%m/%d'),
        instance.exporting_data_sub_stage.__class__.__name__.lower(),
        str(instance.exporting_data_sub_stage_id),
        str(instance.operation),
        filename,
    )


def update_fields(
    entity: 'BaseEntityModel',
    field_values: Dict[str, Any],
    mapping: Dict[str, str]
) -> None:
    """Обновление значений полей сущности по измененным полям модели.

    :param entity: Выгружаемая сущность
    :param field_values: Словарь с измененными данными модели
    :param mapping: Словарь маппинга полей модели к полям сущности
    """
    for model_field, entity_field in mapping.items():
        if model_field in field_values:
            setattr(entity, entity_field, field_values[model_field])


def get_isoformat_timezone():
    """Возвращает временную зонну в ISO представлении."""
    return datetime.now().astimezone().isoformat()[-6:]



from typing import (
    List,
)

from edu_rdm_integration.adapters.caches import (
    WebEduFunctionCacheStorage,
    WebEduRunnerCacheStorage,
)
from educommon.integration_entities.entities import (
    BaseEntity,
)


class BaseExportDataRunnerCacheStorage(WebEduRunnerCacheStorage):
    """
    Базовый кеш помощников ранеров функций выгрузки данных для интеграции с "Региональная витрина данных".
    """


class BaseExportDataFunctionCacheStorage(WebEduFunctionCacheStorage):
    """
    Базовый кеш помощников функций выгрузки данных для интеграции с "Региональная витрина данных".
    """

    def _prepare_entity_instances(self, model_ids) -> List[BaseEntity]:
        """
        Формирование списка объектов сущностей для дальнейшей выгрузки.

        Необходимо переопределить логику формирования объектов у потомков.
        """
        return []

    def __init__(self, *args, model_ids, entities, **kwargs):
        super().__init__(*args, **kwargs)

        self._entity_enum_value = entities[0]

        self.entity_instances = self._prepare_entity_instances(model_ids=model_ids)



from abc import (
    ABCMeta,
)
from typing import (
    Optional,
)

from edu_rdm_integration.adapters.caches import (
    WebEduFunctionCacheStorage,
    WebEduRunnerCacheStorage,
)
from edu_rdm_integration.adapters.errors import (
    WebEduError,
)
from edu_rdm_integration.adapters.functions import (
    WebEduFunction,
    WebEduLazySavingPredefinedQueueFunction,
)
from edu_rdm_integration.adapters.helpers import (
    WebEduFunctionHelper,
    WebEduRunnerHelper,
)
from edu_rdm_integration.adapters.managers import (
    WebEduRunnerManager,
)
from edu_rdm_integration.adapters.presenters import (
    WebEduResultPresenter,
)
from edu_rdm_integration.adapters.results import (
    WebEduFunctionResult,
    WebEduRunnerResult,
)
from edu_rdm_integration.adapters.runners import (
    WebEduRunner,
)
from edu_rdm_integration.adapters.validators import (
    WebEduFunctionValidator,
    WebEduRunnerValidator,
)
from function_tools.strategies import (
    FunctionImplementationStrategy,
)


class WebEduFunctionImplementationStrategy(FunctionImplementationStrategy, metaclass=ABCMeta):
    """
    Базовый класс стратегии реализации функции продуктов Образования.
    """

    def _prepare_manager_class(self):
        """
        Устанавливает класс менеджера.
        """
        self._manager_class = WebEduRunnerManager

    def _prepare_runner_class(self):
        """
        Устанавливает класс ранера.
        """
        self._runner_class = WebEduRunner

    def _prepare_function_class(self):
        """
        Устанавливает класс Функции.
        """
        self._function_class = WebEduFunction

    def _prepare_runner_helper_class(self):
        """
        Устанавливает класс помощника ранера.
        """
        self._runner_helper_class = WebEduRunnerHelper

    def _prepare_function_helper_class(self):
        """
        Устанавливает класс помощника функции.
        """
        self._function_helper_class = WebEduFunctionHelper

    def _prepare_runner_validator_class(self):
        """
        Устанавливает класс валидатора ранера.
        """
        self._runner_validator_class = WebEduRunnerValidator

    def _prepare_function_validator_class(self):
        """
        Устанавливает класс валидатора Функции.
        """
        self._function_validator_class = WebEduFunctionValidator

    def _prepare_runner_cache_storage_class(self):
        """
        Устанавливает класс хранилища кешей ранера.
        """
        self._runner_cache_storage_class = WebEduRunnerCacheStorage

    def _prepare_function_cache_storage_class(self):
        """
        Устанавливает класс хранилища кешей Функции.
        """
        self._function_cache_storage_class = WebEduFunctionCacheStorage

    def _prepare_error_class(self):
        """
        Устанавливает класс ошибки.
        """
        self._error_class = WebEduError

    def _prepare_runner_result_class(self):
        """
        Устанавливает класс результата.
        """
        self._runner_result_class = WebEduRunnerResult

    def _prepare_function_result_class(self):
        """
        Устанавливает класс результата.
        """
        self._function_result_class = WebEduFunctionResult

    def _prepare_result_presenter_class(self):
        """
        Устанавливает класс презентера результата.
        """
        self._result_presenter_class = WebEduResultPresenter


class WebEduSyncBaseRunnerLazySavingPredefinedQueueFunctionImplementationStrategy(
    WebEduFunctionImplementationStrategy,
    metaclass=ABCMeta
):
    """
    Стратегия создания функции с отложенным сохранением и предустановленной очередью объектов на сохранение продуктов Образования.
    """

    def _prepare_key(self) -> str:
        """
        Возвращает уникальный идентификатор стратегии создания функции.
        """
        return 'WEB_EDU_SYNC_LAZY_SAVING_FUNCTION'

    def _prepare_title(self) -> str:
        """
        Возвращает название стратегии создания функции.
        """
        return (
            'Стратегия создания функции с отложенным сохранением и предустановленной очередью объектов на сохранение '
            'продуктов Образования. Сохранение производится после удачной работы функции'
        )

    @classmethod
    def _prepare_function_template_name(cls) -> Optional[str]:
        """
        Формирование названия шаблона создания функции.
        """
        return 'function_sync_template'

    def _prepare_function_class(self):
        """
        Устанавливает класс Функции.
        """
        self._function_class = WebEduLazySavingPredefinedQueueFunction



from typing import (
    Any,
    Dict,
    Optional,
)

from edu_rdm_integration.adapters.caches import (
    WebEduEntityCache,
)


class WebEduEntityCacheExtended(WebEduEntityCache):
    """
    Расширенный класс для кэша сущностей ЭШ.

    Добавлена поддержка annotation и exclude
    """

    def __init__(
        self,
        additional_exclude_params: Optional[Dict[str, Any]] = None,
        annotated_fields: Optional[Dict] = None,
        *args,
        **kwargs
    ):
        self._additional_exclude_params = additional_exclude_params or {}
        self._annotated_fields = annotated_fields

        super().__init__(*args, **kwargs)

    def _prepare_entities(self):
        """
        Получение выборки объектов модели по указанными параметрам.
        """
        self._entities = self._actual_entities_queryset.filter(
            **self._additional_filter_params,
        ).exclude(**self._additional_exclude_params)

        if self._annotated_fields:
            self._entities = self._entities.annotate(**self._annotated_fields)

        if self._only_fields:
            self._entities = self._entities.only(*self._only_fields)

        self._entities = self._entities.distinct()



from sys import (
    stdout,
)


def m3_edu_function_tools_after_migrate_receiver(sender, **kwargs):
    """
    Действия выполняемые после прогона миграций.
    """
    from function_tools.management.commands.register_entities import (
        EntityRegistrar,
    )

    try:
        registrar = EntityRegistrar(logger=stdout)
        registrar.run()
    except Exception as e:
        stdout.write(
            f'Register function_tools_entities exception. ----- START IGNORING EXCEPTION\n{e}\n----- '
            f'FINISH IGNORING EXCEPTION\n'
        )
